from dataclasses import dataclass, field
from datetime import date
from enum import Enum
from typing import Optional


class State(Enum):
    OPEN = " "
    CHECKED = "x"
    ONGOING = "@"
    OBSOLETE = "~"


class LineType(Enum):
    ERROR = 0
    GROUP_TITLE = 1
    ITEM_START = 2
    ITEM_CONTINUATION = 3
    WHITESPACE = 4


@dataclass()
class Group:
    title: Optional[str]
    items: list = field(default_factory=list)


@dataclass()
class Item:
    state: State
    description: str
    tags: set = field(default_factory=set)
    priority: int = 0
    due_date: Optional[date] = None


@dataclass(frozen=True)
class Tag:
    value: Optional[str] = None
    key: Optional[str] = None


@dataclass()
class Line:
    line_type: LineType
    text: str


class ParseError(Exception):
    pass



from datetime import datetime, timedelta


def parse_ymd(format_, date_):
    # exact date
    return datetime.strptime(date_, format_).date()


def parse_ym(format_, date_):
    # last day of month

    # fmt: off
    def last_day_of_month(d):
        return (
            (d.replace(day=28) + timedelta(days=4))
            .replace(day=1) + timedelta(days=-1)
        )
    # fmt: on

    return last_day_of_month(datetime.strptime(date_, format_).date())


def parse_y(date_):
    # last day of year
    return datetime.strptime(date_ + "-12-31", "%Y-%m-%d").date()


def parse_w(format_, date_):
    # last day of week
    return datetime.strptime(date_ + " 0", format_ + " %w").date()


def parse_q(date_):
    # last day of quarter
    quarters = {
        "1": "-03-31",
        "2": "-06-30",
        "3": "-09-30",
        "4": "-12-31",
    }
    try:
        return datetime.strptime(date_[:4] + quarters[date_[6]], "%Y-%m-%d").date()
    except KeyError:
        raise ValueError("Q must be in the range 1..4")



import re
import sys
import unicodedata

_TAG_START = r"\#"
_TAG_KEY = r"([\w\d_-]+)"


_VALUE_ONLY_END = r"(?=[^\w\d_-]|$)"
VALUE_ONLY_TAG_PATTERN = re.compile(_TAG_START + _TAG_KEY + _VALUE_ONLY_END, re.U)


DOUBLE_QUOTED_KEY_VALUE_TAG_PATTERN = re.compile(
    _TAG_START + _TAG_KEY + r"=(\".*?\")", re.U
)


SINGLE_QUOTED_KEY_VALUE_TAG_PATTERN = re.compile(
    _TAG_START + _TAG_KEY + r"=('.*?')", re.U
)


_UNICODE_PUNCTUATION = [
    chr(i)
    for i in range(sys.maxunicode)
    if unicodedata.category(chr(i)).startswith("P")
]
_TAG_UNICODE_PUNCTUATION = "".join([c for c in _UNICODE_PUNCTUATION])
_UNQUOTED_KEY_VALUE_END = r"(?=[ " + _TAG_UNICODE_PUNCTUATION + r"]|$)"
UNQUOTED_KEY_VALUE_TAG_PATTERN = re.compile(
    _TAG_START + _TAG_KEY + r"=([\w\d_-]*)" + _UNQUOTED_KEY_VALUE_END, re.U
)



import sys
import unicodedata
from functools import partial

from stage_left import date_parsers

_DATE_PARSERS = {
    r"(\d{4}-\d{2}-\d{2})": partial(date_parsers.parse_ymd, "%Y-%m-%d"),
    r"(\d{4}/\d{2}/\d{2})": partial(date_parsers.parse_ymd, "%Y/%m/%d"),
    r"(\d{4}-\d{2})": partial(date_parsers.parse_ym, "%Y-%m"),
    r"(\d{4}/\d{2})": partial(date_parsers.parse_ym, "%Y/%m"),
    r"(\d{4}-W\d{2})": partial(date_parsers.parse_w, "%Y-W%W"),
    r"(\d{4}/W\d{2})": partial(date_parsers.parse_w, "%Y/W%W"),
    r"(\d{4}-Q\d{1})": date_parsers.parse_q,
    r"(\d{4}/Q\d{1})": date_parsers.parse_q,
    r"(\d{4})": date_parsers.parse_y,
}


_UNICODE_PUNCTUATION = [
    chr(i)
    for i in range(sys.maxunicode)
    if unicodedata.category(chr(i)).startswith("P")
]
_DATE_UNICODE_PUNCTUATION = "".join(
    [c for c in _UNICODE_PUNCTUATION if c not in ["-", "/"]]
)

_DUE_DATE_START = r"([ " + _DATE_UNICODE_PUNCTUATION + r"]|^)-> "
_DUE_DATE_END = r"(?=[ " + _DATE_UNICODE_PUNCTUATION + r"]|$)"

DUE_DATE_PATTERNS = {
    _DUE_DATE_START + regex + _DUE_DATE_END: parser
    for regex, parser in _DATE_PARSERS.items()
}



import re

from stage_left.date_patterns import DUE_DATE_PATTERNS
from stage_left.tag_patterns import (
    DOUBLE_QUOTED_KEY_VALUE_TAG_PATTERN,
    SINGLE_QUOTED_KEY_VALUE_TAG_PATTERN,
    UNQUOTED_KEY_VALUE_TAG_PATTERN,
    VALUE_ONLY_TAG_PATTERN,
)
from stage_left.types import Group, Item, Line, LineType, ParseError, State, Tag

ALLOWED_ITEM_STARTS = [f"[{state.value}] " for state in State]
ALLOWED_ITEM_LINES = [s.strip() for s in ALLOWED_ITEM_STARTS]


def classify_line(line, prev_line_type, index):
    line_type = LineType.ERROR

    line_start = line[0:4]
    if line_start in ALLOWED_ITEM_STARTS or line in ALLOWED_ITEM_LINES:
        line_type = LineType.ITEM_START
    elif len(line.strip()) == 0:
        line_type = LineType.WHITESPACE
    elif line_start == "    " and prev_line_type in [
        LineType.ITEM_START,
        LineType.ITEM_CONTINUATION,
    ]:
        line_type = LineType.ITEM_CONTINUATION
    elif (
        line_start.lstrip() == line_start
        and not line_start.startswith("[")
        and (prev_line_type == LineType.WHITESPACE or not prev_line_type)
    ):
        line_type = LineType.GROUP_TITLE

    if line_type == LineType.ERROR:
        raise ParseError(f"Error on line {index}:\n{line}")

    return Line(line_type=line_type, text=line)


def classify_lines(lines):
    classified_lines = []
    for index, line in enumerate(lines):
        prev_line_type = None
        if index > 0:
            prev_line_type = classified_lines[index - 1].line_type
        classified_lines.append(classify_line(line, prev_line_type, index))
    return classified_lines


def parse_tags(text):
    temp = text
    tags = set()

    # #key='value'
    matches = re.findall(SINGLE_QUOTED_KEY_VALUE_TAG_PATTERN, temp)
    temp = re.sub(SINGLE_QUOTED_KEY_VALUE_TAG_PATTERN, "", temp)
    for match in matches:
        tags.add(Tag(key=match[0], value=match[1].lstrip("'").rstrip("'")))

    # #key="value"
    matches = re.findall(DOUBLE_QUOTED_KEY_VALUE_TAG_PATTERN, temp)
    temp = re.sub(DOUBLE_QUOTED_KEY_VALUE_TAG_PATTERN, "", temp)
    for match in matches:
        tags.add(Tag(key=match[0], value=match[1].lstrip('"').rstrip('"')))

    # #key=value
    matches = re.findall(UNQUOTED_KEY_VALUE_TAG_PATTERN, temp)
    temp = re.sub(UNQUOTED_KEY_VALUE_TAG_PATTERN, "", temp)
    for match in matches:
        tags.add(Tag(key=match[0], value=match[1]))

    # #value
    matches = re.findall(VALUE_ONLY_TAG_PATTERN, temp)
    temp = re.sub(VALUE_ONLY_TAG_PATTERN, "", temp)
    for match in matches:
        tags.add(Tag(value=match))

    return tags


def parse_priority(text):
    head, *_ = text.split(" ")

    if len(head.strip("!.")) > 0:
        return 0

    if len(head) > 0 and head[0] == ".":
        # left-padded
        stripped = head.lstrip(".")
        if "." in stripped:
            return 0
        return stripped.count("!")

    if len(head) > 0 and head[-1] == ".":
        # right-padded
        stripped = head.rstrip(".")
        if "." in stripped:
            return 0
        return stripped.count("!")

    if "." in head:
        # neither left-padded nor night-padded
        # but contains padding character (invalid)
        return 0

    return head.count("!")


def parse_due_date(text):
    for regex, parser in DUE_DATE_PATTERNS.items():
        matches = re.search(regex, text)
        if matches:
            try:
                return parser(matches[2])
            except ValueError as e:
                raise ParseError(str(e)) from e
    return None


def parse_item(lines):
    line, *tail = lines
    text = line.text[4:]
    for continuation_line in tail:
        if continuation_line.line_type != LineType.ITEM_CONTINUATION:
            break
        text += "\n" + continuation_line.text[4:]

    tags = parse_tags(text)
    priority = parse_priority(line.text[4:])
    due_date = parse_due_date(text)
    return Item(
        state=State(line.text[1]),
        description=text,
        tags=tags,
        priority=priority,
        due_date=due_date,
    )


def parse_lines(lines):
    groups = []
    current_group = None

    for index, line in enumerate(lines):
        if (
            line.line_type == LineType.WHITESPACE
            or line.line_type == LineType.ITEM_CONTINUATION
        ):
            continue

        if line.line_type == LineType.GROUP_TITLE:
            current_group = Group(title=line.text)
            groups.append(current_group)
            continue
        elif line.line_type == LineType.ITEM_START and index == 0:
            current_group = Group(title="")
            groups.append(current_group)
        elif (
            line.line_type == LineType.ITEM_START
            and index > 0
            and lines[index - 1].line_type == LineType.WHITESPACE
        ):
            current_group = Group(title="")
            groups.append(current_group)

        if line.line_type == LineType.ITEM_START:
            current_group.items.append(parse_item(lines[index:]))

    return groups


def parse_text(text):
    """Parse a [x]it string

    Args:
        text (str): String containing a todo list in [x]it format

    Returns:
        list[Group]

    Raises:
        ParseError
    """
    return parse_lines(classify_lines(text.splitlines()))


def parse_file(fp):
    """Parse a [x]it file

    Args:
        fp (IO): File-like object containing a todo list in [x]it format

    Returns:
        list[Group]

    Raises:
        ParseError
    """
    return parse_text(fp.read())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stage_left']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'stage-left',
    'version': '0.1.1',
    'description': '',
    'long_description': '# stage-left\n\n[![Run tests](https://github.com/chris48s/stage-left/actions/workflows/test.yml/badge.svg?branch=master)](https://github.com/chris48s/stage-left/actions/workflows/test.yml)\n[![codecov](https://codecov.io/gh/chris48s/stage-left/branch/master/graph/badge.svg?token=XS70M8EPCT)](https://codecov.io/gh/chris48s/stage-left)\n[![PyPI Version](https://img.shields.io/pypi/v/stage-left.svg)](https://pypi.org/project/stage-left/)\n![License](https://img.shields.io/pypi/l/stage-left.svg)\n![Python Compatibility](https://img.shields.io/badge/dynamic/json?query=info.requires_python&label=python&url=https%3A%2F%2Fpypi.org%2Fpypi%2Fstage-left%2Fjson)\n![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)\n\n[[x]it!](https://xit.jotaen.net/) is a plain-text file format for todos and check lists. Stage-left parses [x]it! documents into rich python objects.\n\n## Installation\n\n```\npip install stage-left\n```\n\n## Usage\n\n### Parse checklist from file\n\n```py\nfrom stage_left import parse_file, ParseError\n\nwith open("/path/to/checklist.xit") as fp:\n    try:\n        checklist = parse_file(fp)\n        print(checklist)\n    except ParseError:\n        raise\n```\n\n### Parse checklist from string\n\n```py\nfrom stage_left import parse_text, ParseError\n\ntext = """\n[ ] Open\n[x] Done\n"""\n\ntry:\n    checklist = parse_text(text)\n    print(checklist)\nexcept ParseError:\n    raise\n```\n\n## Implementation notes\n\nDue dates specified using the numbered week syntax e.g: `2022-W01` are parsed assuming Monday is the first day of the week. All days in a new year preceding the first Monday are considered to be in week 0 (`W00`).\n',
    'author': 'chris48s',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/chris48s/stage-left',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
